@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -360,12 +360,11 @@ static int open_flag_perm(
 		return fd;
 	}
 
-#if defined(O_CREATE)
-	if (!(flags & O_CREATE)) {
 #if defined(O_DIRECTORY)
+	if (!(flags & O_CREAT)) {
 		if (flags & O_DIRECTORY) {
-			(void)mkdir(filename);
-		} else
+			(void)mkdir(filename, mode);
+		} else {
 			fd = open_arg3(filename, O_CREAT | O_RDWR, mode, duration, count);
 			if (fd >= 0)
 				(void)close(fd);
@@ -374,8 +373,7 @@ static int open_flag_perm(
 #else
 	fd = open_arg3(filename, O_CREAT | O_RDWR, mode, duration, count);
 	if (fd >= 0)
-		void)close(fd);
-#endif
+		(void)close(fd);
 #endif
 	fd = open_arg3(filename, flags, mode, duration, count);
 #if defined(O_DIRECTORY)
@@ -1056,7 +1054,7 @@ static int stress_open(const stress_args_t *args)
 					goto close_all;
 				}
 
-				idx = stress_mwc32() % SIZEOF_ARRAY(open_funcs);
+				idx = stress_mwc32modn(SIZEOF_ARRAY(open_funcs));
 				fds[i] = open_funcs[idx](args, mypid, &duration, &count);
 
 				if (fds[i] >= 0)
@@ -1114,7 +1112,7 @@ close_all:
 	}
 
 	rate = (count > 0.0) ? duration / count: 0.0;
-	stress_misc_stats_set(args->misc_stats, 0, "nanosecs per open", rate * 1000000000);
+	stress_metrics_set(args, 0, "nanosecs per open", rate * 1000000000);
 
 	return EXIT_SUCCESS;
 }
